/*******************************************************************************
  Snes9x - Portable Super Nintendo Entertainment System (TM) emulator.
 
  (c) Copyright 1996 - 2003 Gary Henderson (gary.henderson@ntlworld.com) and
                            Jerremy Koot (jkoot@snes9x.com)

  (c) Copyright 2002 - 2003 Matthew Kendora and
                            Brad Jorsch (anomie@users.sourceforge.net)
 

                      
  C4 x86 assembler and some C emulation code
  (c) Copyright 2000 - 2003 zsKnight (zsknight@zsnes.com),
                            _Demo_ (_demo_@zsnes.com), and
                            Nach (n-a-c-h@users.sourceforge.net)
                                          
  C4 C++ code
  (c) Copyright 2003 Brad Jorsch

  DSP-1 emulator code
  (c) Copyright 1998 - 2003 Ivar (ivar@snes9x.com), _Demo_, Gary Henderson,
                            John Weidman (jweidman@slip.net),
                            neviksti (neviksti@hotmail.com), and
                            Kris Bleakley (stinkfish@bigpond.com)
 
  DSP-2 emulator code
  (c) Copyright 2003 Kris Bleakley, John Weidman, neviksti, Matthew Kendora, and
                     Lord Nightmare (lord_nightmare@users.sourceforge.net

  OBC1 emulator code
  (c) Copyright 2001 - 2003 zsKnight, pagefault (pagefault@zsnes.com)
  Ported from x86 assembler to C by sanmaiwashi

  SPC7110 and RTC C++ emulator code
  (c) Copyright 2002 Matthew Kendora with research by
                     zsKnight, John Weidman, and Dark Force

  S-RTC C emulator code
  (c) Copyright 2001 John Weidman
  
  Super FX x86 assembler emulator code 
  (c) Copyright 1998 - 2003 zsKnight, _Demo_, and pagefault 

  Super FX C emulator code 
  (c) Copyright 1997 - 1999 Ivar and Gary Henderson.



 
  Specific ports contains the works of other authors. See headers in
  individual files.
 
  Snes9x homepage: http://www.snes9x.com
 
  Permission to use, copy, modify and distribute Snes9x in both binary and
  source form, for non-commercial purposes, is hereby granted without fee,
  providing that this license information and copyright notice appear with
  all copies and any derived work.
 
  This software is provided 'as-is', without any express or implied
  warranty. In no event shall the authors be held liable for any damages
  arising from the use of this software.
 
  Snes9x is freeware for PERSONAL USE only. Commercial users should
  seek permission of the copyright holders first. Commercial use includes
  charging money for Snes9x or software derived from Snes9x.
 
  The copyright holders request that bug fixes and improvements to the code
  should be forwarded to them so everyone can benefit from the modifications
  in future versions.
 
  Super NES and Super Nintendo Entertainment System are trademarks of
  Nintendo Co., Limited and its subsidiary companies.
*******************************************************************************/

/*******************************************************************************
  Snes9x custom (Snes9x for Mac OS) by zones
  
  This build was started from John Stiles' port.
*******************************************************************************/

#include "snes9x.h"
#include "display.h"
#include "gfx.h"
#include "memmap.h"
#include "ppu.h"
#include "soundux.h"

#include <DrawSprocket.h>
#include "os9-prefix.h"
#include "os9-2xsai.h"
#include "os9-blit.h"
#include "os9-cheatfinder.h"
#include "os9-os.h"
#include "os9-screenshot.h"
#include "os9-render.h"

#undef  CALL_NOT_IN_CARBON
#define CALL_NOT_IN_CARBON 1
#include <QD3DAcceleration.h>
#include <RAVE.h>
#include <RAVESystem.h>
#undef  CALL_NOT_IN_CARBON
#define CALL_NOT_IN_CARBON 0

int	imageWidth  = 0,
	imageHeight = 0,
	imageDepth  = 16;

static OSStatus BlitMPTask(void *);
static OSStatus PrepareMPBlit(void);
static void S9xInitFullScreen(void);
static void S9xDeinitFullScreen(void);
static void S9xInitDirectFullScreen(void);
static void S9xDeinitDirectFullScreen(void);
static void S9xInitDrawSprocket(void);
static void S9xDeinitDrawSprocket(void);
static void S9xInitRAVEFullScreen(void);
static void S9xDeinitRAVEFullScreen(void);
static void S9xPutImageDirect(int, int);
static void S9xPutImageRAVE(int, int);
static inline void RAVESetVertices(TQAVTexture[], float, float, float, float, float, float, float, float);

#define	BYTES_PER_PIXEL	2

enum
{
	kMPBlitFrame = 1,
	kMPBlitDone
};

typedef void (* Blitter) (register unsigned char *, register unsigned char *, register long, register Rect *, register Rect *);

typedef struct
{
	Blitter	blitFn;
	Rect	srcRect;
	Rect	copyRect;
	long	dstRowBytes;
	uint8	*gfxBuffer;
	uint8	*dstPtr;
}	MPData;

static uint8				*gfxScreenA,
							*gfxScreenB,
							*snesScreenA,
							*snesScreenB;

static unsigned char 		*gMacScreenPtr,
							*gWinScreenPtr;
static long					gMacRowBytes;

static Blitter				BlitFn256x240     = nil,
							BlitFn512x240     = nil,
							BlitFn512x480     = nil;

static MPTaskID				taskID            = nil;
static MPQueueID			notificationQueue = nil,
							taskQueue         = nil;
static MPSemaphoreID		readySemaphore    = nil;
static MPData 				*mpDataBuffer     = nil;

static CGrafPtr				frontBuffer       = nil;
static DSpContextReference	dspContext        = nil;

static TQARect				raverect;
static TQADevice			ravedevice;
static TQAEngine			*raveengine       = nil;
static TQADrawContext  		*ravecontext      = nil;
static TQAVTexture			ravevertex[4];

static Boolean				raveavailable     = false;

void InitGraphics(void)
{
	short	safebuffersize = 520 * 520 * BYTES_PER_PIXEL - 512 * 512 * BYTES_PER_PIXEL;
	
	GFX.Pitch      = 512 * BYTES_PER_PIXEL;
	
	snesScreenA    = (uint8 *) calloc(520 * 520 * BYTES_PER_PIXEL, 1);
	snesScreenB    = (uint8 *) calloc(520 * 520 * BYTES_PER_PIXEL, 1);
	
	gfxScreenA     = snesScreenA + (safebuffersize >> 1);
	gfxScreenB     = snesScreenB + (safebuffersize >> 1);
	
	GFX.Screen     = gfxScreenA;
	GFX.SubScreen  = (uint8 *) calloc((GFX.Pitch * 512), 1);
	GFX.ZBuffer    = (uint8 *) calloc((GFX.Pitch >> 1) * IMAGE_HEIGHT, 1);
	GFX.SubZBuffer = (uint8 *) calloc((GFX.Pitch >> 1) * IMAGE_HEIGHT, 1);
    
    if (!snesScreenA || !snesScreenB || !GFX.SubScreen || !GFX.ZBuffer || !GFX.SubZBuffer)
		ExitToShell();
		
	Init_2xSaI(555);
}

void DeinitGraphics(void)
{
	if (snesScreenA)
		free(snesScreenA);
	if (snesScreenB)
		free(snesScreenB);	
	if (GFX.SubScreen)
		free(GFX.SubScreen);
	if (GFX.ZBuffer)
		free(GFX.ZBuffer);
	if (GFX.SubZBuffer)
		free(GFX.SubZBuffer);
}

void InitMP(void)
{		
	if (MPLibraryIsLoaded())
	{
		if (MPProcessorsScheduled() > 1)
			multiprocessor = true;
	}
}

static OSStatus BlitMPTask(void *parameter)
{
	OSStatus	err = noErr;
    long		theCommand, param1, param2;

	#pragma unused (parameter)
	
	//printf("MP: Entered DirectDraw thread.\n");
    
	for (;;)
    {
		err = MPWaitOnQueue(taskQueue, (void **) &theCommand, (void **) &param1, (void **) &param2, kDurationForever);
        if (err)
			break;
               
		if (theCommand == kMPBlitFrame)
		{
			(mpDataBuffer->blitFn) (mpDataBuffer->gfxBuffer, mpDataBuffer->dstPtr, mpDataBuffer->dstRowBytes, &mpDataBuffer->srcRect, &mpDataBuffer->copyRect);
 			MPSignalSemaphore(readySemaphore);
		}
		else
		if (theCommand == kMPBlitDone)
			break;
		else
		{
			err = userCanceledErr;
			break;
		}
	}

	MPFree(mpDataBuffer);
	MPDeleteSemaphore(readySemaphore);
    MPDeleteQueue(taskQueue);
	mpDataBuffer = nil;
	readySemaphore = nil;
	taskQueue = nil;
	
	//printf("MP: Exited DirectDraw thread.\n");

	return err;
}

static OSStatus PrepareMPBlit(void)
{
	OSStatus	err;
	
	mpDataBuffer = (MPData *) MPAllocateAligned(sizeof(MPData), kMPAllocateDefaultAligned, kMPAllocateClearMask);
	if (!mpDataBuffer)
		return memFullErr;
	
	err = MPCreateQueue(&notificationQueue);
	if (err == noErr)
	{
		err = MPCreateQueue(&taskQueue);
		if (err == noErr)
		{
			err = MPCreateBinarySemaphore(&readySemaphore);
			if (err == noErr)
			{
				MPSignalSemaphore(readySemaphore);
				err = MPCreateTask(BlitMPTask, nil, 0, notificationQueue, nil, nil, 0, &taskID);
			}
		}
	}

	return err;
}

void ClearGFXScreen(void)
{
	UInt16	*p, *q, x, y;
	
	p = (UInt16 *) gfxScreenA;
	q = (UInt16 *) gfxScreenB;
	
	for (x = 0; x < 512; x++)
		for (y = 0; y < 512; y++)
			*p++ = *q++ = 0;
			
	BlitUpdateEntireDelta();
	
	if (drawingMethod == kDrawingDirect)
		S9xPutImageDirect(512, 480);
}

static void S9xInitFullScreen(void)
{
	if (((Settings.ControllerOption == SNES_SUPERSCOPE ) ||
		 (Settings.ControllerOption == SNES_JUSTIFIER  ) ||
		 (Settings.ControllerOption == SNES_JUSTIFIER_2)) && scopecursor)
	{
		CursHandle	crsr;

		crsr = GetCursor(128);
		SetCursor(*crsr);
	}
	else
		HideCursor();
		
	S9xInitDrawSprocket();
}

static void S9xDeinitFullScreen(void)
{
	S9xDeinitDrawSprocket();
	
	InitCursor();
}

static void S9xInitDirectFullScreen(void)
{
	GDHandle	monitor;
	Rect		monitorRect, sectionRect, rct;
	size_t		offsetX, offsetY;
	
	if (frontBuffer == nil)
		ExitToShell();
		
	monitor = GetMainDevice();
	
	Point	x = { 0, 0 };
	
	SetPort(frontBuffer);
	LocalToGlobal(&x);
	
	GetPortBounds(frontBuffer, &rct);
	ForeColor(blackColor);
	PaintRect(&rct);
	
	OffsetRect(&rct, x.h, x.v);
	
	glScreenW = rct.right  - rct.left;
	glScreenH = rct.bottom - rct.top;
	
	offsetX = (glScreenW - 512) / 2;
	offsetY = (glScreenH - 480) / 2;

	monitorRect = (**monitor).gdRect;
		
	if ((SectRect(&rct, &monitorRect, &sectionRect)) &&
		(sectionRect.top    == rct.top   ) &&
		(sectionRect.left   == rct.left  ) &&
		(sectionRect.bottom == rct.bottom) &&
		(sectionRect.right  == rct.right ))
	{
		gMacRowBytes  =          GetPixRowBytes((**monitor).gdPMap);
		gMacScreenPtr = (Byte *) GetPixBaseAddr((**monitor).gdPMap);			
		gWinScreenPtr = (Byte *) (gMacScreenPtr + ((rct.top  + offsetY - monitorRect.top ) * gMacRowBytes) +
							 		              ((rct.left + offsetX - monitorRect.left) * 2));
	}
	else
		ExitToShell();

	BlitFn256x240 = doubleSize ? (saiMode ? BlitPix2xSaI16 : (eagleMode ? BlitPixEagle16 : (tvMode ? BlitPixScaledTV16 : (smoothMode ? BlitPixFiltered16 : BlitPixScaled16)))) : BlitPixSmall16;
	BlitFn512x240 = tvMode ? BlitPixHiResTV16 : BlitPixHiRes16;
	BlitFn512x480 = BlitPixSmall16;
	
	BlitUpdateEntireDelta();
	
	if (multiprocessor)
	{
		//printf("MP: Creating DirectDraw thread.\n");
		
		if (noErr != PrepareMPBlit())
			multiprocessor = false;	
	}
}

static void S9xDeinitDirectFullScreen(void)
{
	if (multiprocessor)
	{
		MPNotifyQueue(taskQueue, (void *) kMPBlitDone, 0, 0);
		MPWaitOnQueue(notificationQueue, nil, nil, nil, kDurationForever);
		MPDeleteQueue(notificationQueue);
		notificationQueue = nil;

		//printf("MP: Successfully received terminate signal from DirectDraw thread.\n");
	}
}

static void S9xInitDrawSprocket(void)
{
	if (dspContext == nil)
	{
		OSErr					err;
		DisplayIDType			displayID;
		DSpContextAttributes	attr = { 0 };

		if (!autoRes)
		{
			
			GDHandle	monitor = GetMainDevice();
			attr.displayWidth  = (**monitor).gdRect.right  - (**monitor).gdRect.left;
			attr.displayHeight = (**monitor).gdRect.bottom - (**monitor).gdRect.top;
		}
		else
		{			
			attr.displayWidth  = 640;
			attr.displayHeight = 480;
		}

		attr.colorNeeds          = kDSpColorNeeds_Require;
		attr.backBufferDepthMask = 16;
		attr.displayDepthMask    = 16;
		attr.backBufferBestDepth = 16;
		attr.displayBestDepth    = 16;

		err = DSpStartup();
		if (err == noErr)
		{
			DMGetDisplayIDByGDevice(GetMainDevice(), &displayID, true);
			err = DSpFindBestContextOnDisplayID(&attr, &dspContext, displayID);
			if (err == noErr)
			{
				attr.pageCount = 1;
				err = DSpContext_Reserve(dspContext, &attr);
				if (err == noErr)
				{
					err = DSpContext_SetState(dspContext, kDSpContextState_Active);
					if (err == kDSpConfirmSwitchWarning)
						err = noErr;
				
					if (err == noErr)
						err = DSpContext_GetFrontBuffer(dspContext, &frontBuffer);
				}
			}
		}
	}
}

void S9xDeinitDrawSprocket(void)
{
	if (dspContext != nil)
	{
		OSErr	err;

		err = DSpContext_SetState(dspContext, kDSpContextState_Inactive);
		if (err == noErr)
		{
			err = DSpContext_Release(dspContext);
			if (err == noErr)
				err = DSpShutdown();
		}
		
		dspContext  = nil;
		frontBuffer = nil;
	}
}

static void S9xInitRAVEFullScreen(void)
{
	TQAError		err;
	GDHandle		gdv;
    unsigned long	fast;
    
    raveavailable = false;
    
	gdv = GetMainDevice();
	glScreenW = (**gdv).gdRect.right  - (**gdv).gdRect.left;
	glScreenH = (**gdv).gdRect.bottom - (**gdv).gdRect.top;
	
	/*err = QAEngineEnable(kQAVendor_Apple, kQAEngine_AppleHW);
    if (err != kQANoErr)
    	return;*/
	
	ravedevice.deviceType     = kQADeviceGDevice;
	ravedevice.device.gDevice = gdv;
   
    raveengine = QADeviceGetFirstEngine(&ravedevice);
    if (raveengine == nil)
	    return;

    do
    { 
		err = QAEngineGestalt(raveengine, kQAGestalt_FastFeatures, &fast);
		if ((err == kQANoErr) && (fast & kQAFast_Texture))
			break;
		
		raveengine = QADeviceGetNextEngine(&ravedevice, raveengine);
    }
    while (raveengine != nil);
    
    if (raveengine == nil)
    	raveengine = QADeviceGetFirstEngine(&ravedevice);

	raverect.left   = 0;
	raverect.top    = 0;
	raverect.right  = glScreenW;
	raverect.bottom = glScreenH;
	    
    err = QADrawContextNew(&ravedevice, &raverect, nil, raveengine, kQAContext_NoZBuffer | kQAContext_NoDither | kQAContext_DoubleBuffer, &ravecontext);
    if (err != kQANoErr)
    	return;
    	
    raveavailable = true;

	QASetFloat(ravecontext, kQATag_ColorBG_a, 1.0);
    QASetFloat(ravecontext, kQATag_ColorBG_r, 0.0);
    QASetFloat(ravecontext, kQATag_ColorBG_g, 0.0);
    QASetFloat(ravecontext, kQATag_ColorBG_b, 0.0);
    
	QASetInt(ravecontext, kQATag_BitmapFilter,    smoothMode ? kQATextureFilter_Best : kQATextureFilter_Fast);
    QASetInt(ravecontext, kQATag_TextureFilter,   smoothMode ? kQATextureFilter_Best : kQATextureFilter_Fast);
	QASetInt(ravecontext, kQATag_Antialias,       smoothMode ? kQAAntiAlias_Best     : kQAAntiAlias_Off     );
	QASetInt(ravecontext, kQATag_ZFunction,       kQAZFunction_None      );
	QASetInt(ravecontext, kQATag_PerspectiveZ,    kQAPerspectiveZ_Off    );
	QASetInt(ravecontext, kQATag_BufferComposite, kQABufferComposite_None);

	QARenderStart(ravecontext, nil, nil);
	QARenderEnd(ravecontext, nil);
		
	for (int i = 0; i < 4; i++)
	{
		ravevertex[i].x      = 0.0;
		ravevertex[i].y      = 0.0;
		ravevertex[i].z      = 0.0;
		ravevertex[i].invW   = 1.0;
		ravevertex[i].r      = 1.0;
		ravevertex[i].g      = 1.0;
		ravevertex[i].b      = 1.0;
		ravevertex[i].a      = 1.0;
		ravevertex[i].uOverW = 0.0;
		ravevertex[i].vOverW = 0.0;
		ravevertex[i].kd_r   = 1.0;
		ravevertex[i].kd_g   = 1.0;
		ravevertex[i].kd_b   = 1.0;
		ravevertex[i].ks_r   = 1.0;
		ravevertex[i].ks_g   = 1.0;
		ravevertex[i].ks_b   = 1.0;
	}
}

static void S9xDeinitRAVEFullScreen(void)
{
	if (raveavailable)
	{
		QADrawContextDelete(ravecontext);
		ravecontext = nil;
		
		raveavailable = false;
	} 
}

void S9xInitDisplay(int argc, char **argv)
{
	#pragma unused (argc, argv)
	
	if (directDisplay)
		return;

	S9xInitFullScreen();
		
	switch (drawingMethod)
	{
		case kDrawingDirect:
			gSrcP = 1024;
			Settings.OpenGLEnable = false;
			S9xInitDirectFullScreen();
			break;
			
		case kDrawingOpenGL:	// RAVE
			gSrcP = 1024;
			Settings.OpenGLEnable = false;
			S9xInitRAVEFullScreen();
	}
		
	S9xSetSoundMute(!Settings.APUEnabled);
	Microseconds((UnsignedWide *) &lastFrame);

	directDisplay = true;
}

void S9xDeinitDisplay(void)
{
	if (!directDisplay)
		return;
	
	S9xSetSoundMute(true);
	
	switch (drawingMethod)
	{
		case kDrawingDirect:
			S9xDeinitDirectFullScreen();
			break;
			
		case kDrawingOpenGL:	// RAVE
			S9xDeinitRAVEFullScreen();
	}

	S9xDeinitFullScreen();
	
	Settings.OpenGLEnable = false;
	directDisplay = false;
	
	lastDrawingMethod = drawingMethod;
}

bool8 S9xInitUpdate(void)
{
    GFX.RealPitch = GFX.Pitch = GFX.Pitch2 = 512 * 2;
		
	return true;
}

bool8 S9xDeinitUpdate(int width, int height, bool8 sixteenBit)
{
	if (directDisplay)
		S9xPutImage(width, height);
	
	return true;
}

void S9xPutImage(int width, int height)
{
	if (cfIsWatching)
		CheatFinderDrawWatchAddr();
	
	if (Settings.DisplayFrameRate)
	{
		static int	drawnFrames[60] = { 1,1,1,1,1, 1,1,1,1,1, 1,1,1,1,1, 1,1,1,1,1, 1,1,1,1,1, 1,1,1,1,1,
										1,1,1,1,1, 1,1,1,1,1, 1,1,1,1,1, 1,1,1,1,1, 1,1,1,1,1, 1,1,1,1,1 };
		static int	tableIndex = 0;
		int			frameCalc = 0, count;
		
		drawnFrames[tableIndex] = Settings.SkipFrames;

		if (gFastForward)
			drawnFrames[tableIndex] = (drawnFrames[tableIndex] + (kFastForwardRate / 2)) / kFastForwardRate;

		tableIndex = (tableIndex + 1) % 60;

		for (count = 0; count < 60; count++)
			frameCalc += drawnFrames[count];

		IPPU.DisplayedRenderedFrameCount = (Memory.ROMFramesPerSecond * 60) / frameCalc;
	}

	switch (drawingMethod)
	{
		case kDrawingDirect:
			S9xPutImageDirect(width, height);
			break;

		case kDrawingOpenGL:	// RAVE
			S9xPutImageRAVE(width, height);
	}
}

static void S9xPutImageDirect(int width, int height)
{
	Rect	srcRect = { 0, 0, height, width }, copyRect;
	Blitter	BlitFn  = nil;

	copyRect.top = copyRect.left = 0;
	
	if (width > 256)
	{
		copyRect.right = width;
		
		if (height > 240)
		{
			copyRect.bottom = height;
			BlitFn = BlitFn512x480;
			
			if (height <= 448)
			{
				copyRect.top     = 16;
				copyRect.bottom += 16;
			}
		}
		else
		{			
			copyRect.bottom = height * 2;
			BlitFn = BlitFn512x240;
			
			if (height <= 224)
			{
				copyRect.top     = 16;
				copyRect.bottom += 16;
			}
			else
			if (height == 239)
			{
				copyRect.top     = 1;
				copyRect.bottom += 1;
			}
		}
	}
	else
	{
		if (doubleSize)
		{
			copyRect.right =   width * 2;
			copyRect.bottom = height * 2;
			
			if (height <= 224)
			{
				copyRect.top     = 16;
				copyRect.bottom += 16;
			}
			else
			if (height == 239)
			{
				copyRect.top     = 1;
				copyRect.bottom += 1;
			}
		}
		else
		{
			copyRect.top    = 120;
			copyRect.left   = 128;
			copyRect.bottom = 120 + height;
			copyRect.right  = 128 + width;
			
			if (height <= 224)
			{
				copyRect.top    += 8;
				copyRect.bottom += 8;
			}				
		}

		BlitFn = BlitFn256x240;
	}

	if (multiprocessor)
	{
		MPWaitOnSemaphore(readySemaphore, kDurationForever);

		mpDataBuffer->blitFn      = BlitFn;
		mpDataBuffer->srcRect     = srcRect;
		mpDataBuffer->copyRect    = copyRect;
		mpDataBuffer->gfxBuffer   = GFX.Screen;
		mpDataBuffer->dstPtr      = gWinScreenPtr;
		mpDataBuffer->dstRowBytes = gMacRowBytes;
				
		if ((width <= 256) && ((imageWidth != width) || (imageHeight != height)))
		{
			if (eagleMode || saiMode)
				BlitClearExtendArea(GFX.Screen, &srcRect);
			else
				BlitUpdateEntireDelta();
		}
		
		MPNotifyQueue(taskQueue, (void *) kMPBlitFrame, 0, 0);
		
		GFX.Screen = ((GFX.Screen == gfxScreenA) ? gfxScreenB : gfxScreenA);
	}
	else
	{
		if ((width <= 256) && ((imageWidth != width) || (imageHeight != height)))
		{
			if (eagleMode || saiMode)
				BlitClearExtendArea(GFX.Screen, &srcRect);
			else
				BlitUpdateEntireDelta();
		}
	
		BlitFn(GFX.Screen, gWinScreenPtr, gMacRowBytes, &srcRect, &copyRect);		
	}
	
	//imageDepth  = 16;
	imageWidth  = width;
	imageHeight = height;
}

static void S9xPutImageRAVE(int width, int height)
{
	TQAError	err;
	TQATexture	*texture[4];
	TQAImage	textureimage;
	int			across, down, count = -1;
	int			screenTop    = scopeViewInfo.globalTop,
				screenLeft   = scopeViewInfo.globalLeft,
				screenWidth  = scopeViewInfo.width,
				screenHeight = scopeViewInfo.height;

	if (raveavailable)
	{
		//imageDepth  = 16;
		imageWidth  = width;
		imageHeight = height;
		
		GetGameScreenRect();

		textureimage.width    = 256;
		textureimage.height   = 256;
		textureimage.rowBytes = 1024;

		QARenderStart(ravecontext, nil, nil);
		
		if (width > 256)
		{
			if (height > 256)
			{
				for (down = 0; down < 2; down++)
				{
					for (across = 0; across < 2; across++)
					{
						count++;
						textureimage.pixmap = GFX.Screen + BYTES_PER_PIXEL * ((down * ((height >> 1) * 512)) + (across * (width >> 1)));

						err = QATextureNew(raveengine, kQATexture_None, kQAPixel_RGB16, &textureimage, &texture[count]);
						if (err == kQANoErr)
						{
							err = QATextureDetach(raveengine, texture[count]);
							if (err == kQANoErr)
							{
								QASetPtr(ravecontext, kQATag_Texture, texture[count]);
								RAVESetVertices(ravevertex, screenTop, screenLeft, screenTop + (screenHeight >> 1), screenLeft + (screenWidth >> 1), 1.0 - ((float) height / 512.0), 0.0, 1.0, 1.0);
								QADrawVTexture(ravecontext, 4, kQAVertexMode_Fan, ravevertex, nil);
							}
						}
						
						screenLeft += (screenWidth >> 1);
					}
					
					screenLeft  = scopeViewInfo.globalLeft;
					screenTop  += (screenHeight >> 1);
				}
			}
			else
			{
				for (across = 0; across < 2; across++)
				{
					count++;
					textureimage.pixmap = GFX.Screen + BYTES_PER_PIXEL * (across * (width >> 1));

					err = QATextureNew(raveengine, kQATexture_None, kQAPixel_RGB16, &textureimage, &texture[count]);
					if (err == kQANoErr)
					{
						err = QATextureDetach(raveengine, texture[count]);
						if (err == kQANoErr)
						{
							QASetPtr(ravecontext, kQATag_Texture, texture[count]);
							RAVESetVertices(ravevertex, screenTop, screenLeft, screenTop + screenHeight, screenLeft + (screenWidth >> 1), 1.0 - ((float) height / 256.0), 0.0, 1.0, 1.0);
							QADrawVTexture(ravecontext, 4, kQAVertexMode_Fan, ravevertex, nil);
						}
					}
					
					screenLeft += (screenWidth >> 1);
				}
			}
		}
		else
		{
			count++;
			textureimage.pixmap = GFX.Screen;

			err = QATextureNew(raveengine, kQATexture_None, kQAPixel_RGB16, &textureimage, &texture[count]);
			if (err == kQANoErr)
			{
				err = QATextureDetach(raveengine, texture[count]);
				if (err == kQANoErr)
				{
					QASetPtr(ravecontext, kQATag_Texture, texture[count]);
					RAVESetVertices(ravevertex, screenTop, screenLeft, screenTop + screenHeight, screenLeft + screenWidth, 1.0 - ((float) height / 256.0), 0.0, 1.0, 1.0);
					QADrawVTexture(ravecontext, 4, kQAVertexMode_Fan, ravevertex, nil);
				}
			}
		}
		
		QARenderEnd(ravecontext, nil);
		
		while (count >= 0)
			QATextureDelete(raveengine, texture[count--]);
	}
}

static inline void RAVESetVertices(TQAVTexture v[], float top, float left, float bottom, float right, float Ttop, float Tleft, float Tbottom, float Tright)
{
	const float RAVE_DELTA = 0.0001f;
	const float TEX_ADJUST = 0.002f;

	v[0].x      = left    + RAVE_DELTA;
	v[0].y      = top     + RAVE_DELTA;
	v[0].uOverW = Tleft   + RAVE_DELTA + TEX_ADJUST;
	v[0].vOverW = Tbottom - RAVE_DELTA - TEX_ADJUST;

	v[1].x      = left    + RAVE_DELTA;
	v[1].y      = bottom  - RAVE_DELTA;
	v[1].uOverW = Tleft   + RAVE_DELTA + TEX_ADJUST;
	v[1].vOverW = Ttop    + RAVE_DELTA + TEX_ADJUST;

	v[2].x      = right   - RAVE_DELTA;
	v[2].y      = bottom  - RAVE_DELTA;
	v[2].uOverW = Tright  - RAVE_DELTA - TEX_ADJUST;
	v[2].vOverW = Ttop    + RAVE_DELTA + TEX_ADJUST;

	v[3].x      = right   - RAVE_DELTA;
	v[3].y      = top     + RAVE_DELTA;
	v[3].uOverW = Tright  - RAVE_DELTA - TEX_ADJUST;
	v[3].vOverW = Tbottom - RAVE_DELTA - TEX_ADJUST;
}

void S9xTextMode(void)
{
	return;
}

void S9xGraphicsMode(void)
{
	return;
}

void S9xSetPalette(void) 
{
	return;
}

void SetInfoDlgColor(unsigned char r, unsigned char g, unsigned char b)
{
	#pragma unused (r, g, b)
	
	return;
}
