/*******************************************************************************
  Snes9x - Portable Super Nintendo Entertainment System (TM) emulator.
 
  (c) Copyright 1996 - 2003 Gary Henderson (gary.henderson@ntlworld.com) and
                            Jerremy Koot (jkoot@snes9x.com)

  (c) Copyright 2002 - 2003 Matthew Kendora and
                            Brad Jorsch (anomie@users.sourceforge.net)
 

                      
  C4 x86 assembler and some C emulation code
  (c) Copyright 2000 - 2003 zsKnight (zsknight@zsnes.com),
                            _Demo_ (_demo_@zsnes.com), and
                            Nach (n-a-c-h@users.sourceforge.net)
                                          
  C4 C++ code
  (c) Copyright 2003 Brad Jorsch

  DSP-1 emulator code
  (c) Copyright 1998 - 2003 Ivar (ivar@snes9x.com), _Demo_, Gary Henderson,
                            John Weidman (jweidman@slip.net),
                            neviksti (neviksti@hotmail.com), and
                            Kris Bleakley (stinkfish@bigpond.com)
 
  DSP-2 emulator code
  (c) Copyright 2003 Kris Bleakley, John Weidman, neviksti, Matthew Kendora, and
                     Lord Nightmare (lord_nightmare@users.sourceforge.net

  OBC1 emulator code
  (c) Copyright 2001 - 2003 zsKnight, pagefault (pagefault@zsnes.com)
  Ported from x86 assembler to C by sanmaiwashi

  SPC7110 and RTC C++ emulator code
  (c) Copyright 2002 Matthew Kendora with research by
                     zsKnight, John Weidman, and Dark Force

  S-RTC C emulator code
  (c) Copyright 2001 John Weidman
  
  Super FX x86 assembler emulator code 
  (c) Copyright 1998 - 2003 zsKnight, _Demo_, and pagefault 

  Super FX C emulator code 
  (c) Copyright 1997 - 1999 Ivar and Gary Henderson.



 
  Specific ports contains the works of other authors. See headers in
  individual files.
 
  Snes9x homepage: http://www.snes9x.com
 
  Permission to use, copy, modify and distribute Snes9x in both binary and
  source form, for non-commercial purposes, is hereby granted without fee,
  providing that this license information and copyright notice appear with
  all copies and any derived work.
 
  This software is provided 'as-is', without any express or implied
  warranty. In no event shall the authors be held liable for any damages
  arising from the use of this software.
 
  Snes9x is freeware for PERSONAL USE only. Commercial users should
  seek permission of the copyright holders first. Commercial use includes
  charging money for Snes9x or software derived from Snes9x.
 
  The copyright holders request that bug fixes and improvements to the code
  should be forwarded to them so everyone can benefit from the modifications
  in future versions.
 
  Super NES and Super Nintendo Entertainment System are trademarks of
  Nintendo Co., Limited and its subsidiary companies.
*******************************************************************************/

/*******************************************************************************
  Snes9x custom (Snes9x for Mac OS) by zones
  
  This build was started from John Stiles' port.
*******************************************************************************/

#include "snes9x.h"
#include "soundux.h"

#include "os9-prefix.h"
#include "os9-os.h"
#include "os9-sound.h"

int		soundBufferAt = -1,
		soundQueued   = 0;
char	*soundBuffer[kMaxSoundBuffers];

static pascal void MacProcessSound(SndChannelPtr, SndCommand *);

static int				soundPlayAt  = -1;
static char				*emptyBuffer = nil;
static SndChannelPtr	sndChannel   = nil;
static ExtSoundHeader	sndHeader;
static SndCallBackUPP	callBackUPP;

void InitMacSound(void)
{
	callBackUPP = NewSndCallBackUPP(MacProcessSound);
}

void DeinitMacSound(void)
{
	DisposeSndCallBackUPP(callBackUPP);
	
	if (sndChannel)
	{
		SndDisposeChannel(sndChannel, true);
		sndChannel = nil;
	}
	
	if (soundBufferAt != -1)
	{
		for (int count = 0; count < kMaxSoundBuffers; count++)
			free(soundBuffer[count]);
		free(emptyBuffer);
	}
}

static pascal void MacProcessSound(SndChannelPtr chan, SndCommand *cmd)
{
	#pragma unused (chan, cmd)
		
	if (so.mute_sound || !Settings.APUEnabled || soundQueued <= 0)
		sndHeader.samplePtr = emptyBuffer;
	else
	{   
		sndHeader.samplePtr = soundBuffer[soundPlayAt];		
		if (++soundPlayAt >= macBufferCount)
			soundPlayAt = 0;
		soundQueued--;
	}
	
	SndCommand buffer   = { bufferCmd, 0, (long) &sndHeader };
	SndDoImmediate(sndChannel, &buffer);

	SndCommand callback = { callBackCmd, 0, 0 };
	SndDoCommand(sndChannel, &callback, true);
}

bool8 S9xOpenSoundDevice(int mode, bool8 stereo, int buffer_size)
{	
	#pragma unused (mode, stereo)
	
	int	count;
	
	if (sndChannel)
	{
		SndDisposeChannel(sndChannel, true);
		sndChannel = nil;
	}

	so.playback_rate = Settings.SoundPlaybackRate;
    so.stereo        = Settings.Stereo;
    so.sixteen_bit   = Settings.SixteenBitSound;
    so.encoded       = false;

	macSoundPitchMaster = macSoundPitch;
	
	so.buffer_size = buffer_size;
	if (so.sixteen_bit)
		so.buffer_size <<= 1;
	if (so.stereo)
		so.buffer_size <<= 1;

  	memset(&sndHeader, 0, sizeof(sndHeader));
	sndHeader.numChannels   = so.stereo ? 2 : 1;
	sndHeader.sampleRate    = so.playback_rate << 16;
	sndHeader.encode        = extSH;
	sndHeader.baseFrequency = kMiddleC;
	sndHeader.numFrames     = buffer_size;
	sndHeader.sampleSize    = so.sixteen_bit ? 16 : 8;
  
	if (soundBufferAt != -1)
	{
		for (count = 0; count < kMaxSoundBuffers; count++)
			free(soundBuffer[count]);
		free(emptyBuffer);
	}
	
	for (count = 0; count < kMaxSoundBuffers; count++)
		soundBuffer[count] = (char *) calloc(so.buffer_size, 1);
	emptyBuffer = (char *) calloc(so.buffer_size, 1);
	
	soundBufferAt = soundPlayAt = soundQueued = 0;
	
    SndNewChannel(&sndChannel, sampledSynth, so.stereo ? initStereo : initMono, callBackUPP);
	
	SndCommand	sndcmd;
	UInt32		volume;
	
	volume = (UInt32) (256.0 * (float) macSoundVolume / 100.0);
	
	sndcmd.cmd = volumeCmd;
    sndcmd.param1 = 0;
    sndcmd.param2 = (volume << 16) | volume;
    SndDoCommand(sndChannel, &sndcmd, true);
	
	sndcmd.cmd = callBackCmd;
    sndcmd.param1 = 0;
    sndcmd.param2 = 0;	
	SndDoCommand(sndChannel, &sndcmd, true);
	
	return true;
}

void S9xGenerateSound(void)
{
	return;
}
